/*
 Copyright 2012, S.S. Papadopulos & Assoc. Inc.

    This file is part of GENIE.

    The GENIE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    GENIE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GENIE.  If not, see<http://www.gnu.org/licenses/>.
*/

// cmuffels, Genie version 1, 2012

#include <iostream>
#include "sendor.h"

using namespace std;

//****************************************************************************
void SENDOR::set_calling_program(CLIENT *_model)
//****************************************************************************
//CTM Apr 2011

/*
      Set calling program to which runs are returned
*/

{
  model=_model;
}

//****************************************************************************
void SENDOR::set_queue_results(list<MODEL_RESULT*> *_results)
//****************************************************************************
//CTM Dec 2010

/*
      Set new run shared with primary thread
*/

{
  results=_results;
}

//****************************************************************************
int SENDOR::execute()
//****************************************************************************
//CTM Apr 2011

/*
      Execute thread process
*/

{

  DWORD tr_Wait,m_Wait;
  bool endthread=false;
  HANDLE trEvents[2];

  list<MODEL_RESULT*>::iterator result;

  try
  {

    // duplicate handles of primary thread
    if(!DuplicateHandle(primary,ph_thComm,GetCurrentProcess(),&h_thComm,
                        0,true,DUPLICATE_SAME_ACCESS)) throw -1;
    if(!DuplicateHandle(primary,pm_thComm,GetCurrentProcess(),&m_thComm,
                        0,true,DUPLICATE_SAME_ACCESS)) throw -2;
    if(!DuplicateHandle(primary,ph_thEnd,GetCurrentProcess(),&h_thEnd,
                        0,true,DUPLICATE_SAME_ACCESS)) throw -3;
    trEvents[0]=h_thComm;
    trEvents[1]=h_thEnd;

    while(1)
    {

      if(endthread)
      {
        if(!ResetEvent(h_thEnd)) throw ERROR_EVENT_RELEASE;
        CloseHandle(h_thEnd);
        break;
      }

      // wait for either the execute or terminate flags to be raised
      tr_Wait=WaitForMultipleObjects(2,trEvents,FALSE,INFINITE);
      switch(tr_Wait)
      {
      case WAIT_FAILED:
        cout<<"Wait failed: "<<GetLastError()<<'\n';
        throw -10;
        break;
      case WAIT_TIMEOUT:
        cout<<"Wait timed out unexpectedly."<<'\n';
        throw -10;
        break;
      case WAIT_ABANDONED_0+0:
        cout<<"Wait on COMM event abandoned."<<'\n';
        throw -10;
        break;
      case WAIT_ABANDONED_0+1:
        cout<<"Wait on END event abandoned."<<'\n';
        throw -10;
        break;
      case WAIT_OBJECT_0+0:
        m_Wait=WaitForSingleObject(m_thComm,INFINITE);
        if(m_Wait!=WAIT_OBJECT_0) throw ERROR_EVENT_WAIT;
        result=results->begin();
        while(result!=results->end())
        {
          if((*result)->send(model->sock))
          {
            (*result)->dealloc();
            delete *result;
            result=results->erase(result);
          }
          else
            ++result;
        }
        if(!ReleaseMutex(m_thComm)) throw ERROR_EVENT_RELEASE;
        break;
      case WAIT_OBJECT_0+1:
        // Terminate thread
        endthread=true;
        break;
      default:
        throw ERROR_EVENT_WAIT;
        break;
      }

    }

    return 1;
  }

  catch(int err)
  {
    switch(err)
    {
    case -1:
      cout<<"\n\n Error duplicating misc. communication flag."<<'\n';
      break;
    case -2:
      cout<<"\n\n Error duplicating misc. shared access flag."<<'\n';
      break;
    case -3:
      cout<<"\n\n Error duplicating termination flag."<<'\n';
      break;
    case ERROR_EVENT_WAIT:
      cout<<"\n\n Error waiting on END or COMM events."<<'\n';
      break;
    case ERROR_EVENT_RELEASE:
      cout<<"\n\n Error releasing COMM mutex."<<'\n';
      break;
    }
    cout<<" *** THREAD :"<<threadid<<'\n';
    return 0;
  }

}
